import asyncio
from functools import partial


def _on_result(future: asyncio.Future, new_future: asyncio.Future=None):
    if not new_future.done():
        exc = future.exception()
        if exc:
            return new_future.set_exception(exc)

        new_future.set_result(future.result())


def copy_future(future: asyncio.Future, new_future: asyncio.Future=None):
    new_future = new_future or asyncio.Future(loop=future._loop)

    handler = partial(_on_result, new_future=new_future)

    future.add_done_callback(handler)
    return new_future


@asyncio.coroutine
def wait(tasks, loop=None):
    loop = loop or asyncio.get_event_loop()
    done = yield from asyncio.gather(*list(tasks), loop=loop)
    return tuple(map(lambda x: x.result() if isinstance(x, asyncio.Future) else x, done))



import asyncio
import pika.channel
from functools import partial
from logging import getLogger
from types import FunctionType
from . import exceptions
from .exchange import Exchange, ExchangeType
from .queue import Queue
from .common import BaseChannel, FutureStore, ConfirmationTypes


log = getLogger(__name__)


class Channel(BaseChannel):
    __slots__ = ('__connection', '__closing', '__confirmations', '__delivery_tag',
                 'loop', '_futures', '__channel',)

    def __init__(self, connection,
                 loop: asyncio.AbstractEventLoop, future_store: FutureStore):

        super().__init__(loop, future_store.get_child())

        self.__channel = None  # type: pika.channel.Channel
        self.__connection = connection
        self.__confirmations = {}
        self.__delivery_tag = 0

    def __str__(self):
        return "{0}".format(self.__channel.channel_number)

    def __repr__(self):
        return '<Channel "%s#%s">' % (self.__connection, self)

    def _on_channel_close(self, channel: pika.channel.Channel, code: int, reason):
        exc = exceptions.ChannelClosed(code, reason)
        log.error("Channel %r closed: %d - %s", channel, code, reason)

        self._futures.reject_all(exc)
        self._closing.set_exception(exc)

    def add_close_callback(self, callback: FunctionType):
        self._closing.add_done_callback(lambda r: callback(r.result()))

    @asyncio.coroutine
    def initialize(self, timeout=None) -> None:
        if self._closing.done():
            raise RuntimeError("Can't initialize closed channel")

        future = self._create_future(timeout=timeout)

        self.__connection._connection.channel(future.set_result)

        channel = yield from future  # type: pika.channel.Channel
        channel.confirm_delivery(self._on_delivery_confirmation)
        channel.add_on_close_callback(self._on_channel_close)

        self.__channel = channel

    def _on_delivery_confirmation(self, method_frame):
        future = self.__confirmations.pop(method_frame.method.delivery_tag, None)

        if not future:
            log.warning(
                "Unknown delivery tag %d for message confirmation \"%s\"",
                method_frame.method.delivery_tag, method_frame.method.NAME)
            return

        try:
            confirmation_type = ConfirmationTypes(method_frame.method.NAME.split('.')[1].lower())

            if confirmation_type == ConfirmationTypes.ACK:
                future.set_result(True)
            elif confirmation_type == ConfirmationTypes.NACK:
                future.set_exception(exceptions.NackError(method_frame))
        except ValueError:
            future.set_exception(RuntimeError('Unknown method frame', method_frame))
        except Exception as e:
            future.set_exception(e)

    @BaseChannel._ensure_channel_is_open
    @asyncio.coroutine
    def declare_exchange(self, name: str, type: ExchangeType = ExchangeType.DIRECT,
                         durable: bool = None, auto_delete: bool = False,
                         arguments: dict = None, timeout: int = None) -> Exchange:

        if auto_delete and durable is None:
            durable = False

        f = self._create_future(timeout=timeout)

        self.__channel.exchange_declare(
            f.set_result,
            name, ExchangeType(type).value, durable=durable,
            auto_delete=auto_delete, arguments=arguments
        )

        yield from f

        exchange = Exchange(
            self.__channel, self._publish, name, type,
            durable=durable, auto_delete=auto_delete, arguments=arguments,
            loop=self.loop, future_store=self._futures.get_child(),
        )

        log.debug("Exchange declared %r", exchange)

        return exchange

    @BaseChannel._ensure_channel_is_open
    @asyncio.coroutine
    def _publish(self, queue_name, routing_key, body, properties, mandatory, immediate):
        while self.__connection.is_closed:
            log.debug("Can't publish message because connection is inactive")
            yield from asyncio.sleep(1, loop=self.loop)

        f = self._create_future()

        try:
            self.__channel.basic_publish(queue_name, routing_key, body, properties, mandatory, immediate)
        except (AttributeError, RuntimeError) as exc:
            log.exception("Failed to send data to client. Conection unexpected closed.")
            self._on_channel_close(self.__channel, -1, exc)
            self.__connection.close(reply_code=500, reply_text="Incorrect state")
        else:
            self.__delivery_tag += 1
            self.__confirmations[self.__delivery_tag] = f

        return (yield from f)

    @BaseChannel._ensure_channel_is_open
    @asyncio.coroutine
    def declare_queue(self, name: str = '', *, durable: bool = None, exclusive: bool = False,
                      auto_delete: bool = False, arguments: dict = None, timeout: int = None) -> Queue:

        if auto_delete and durable is None:
            durable = False

        queue = Queue(
            self.loop, self._futures.get_child(), self.__channel, name,
            durable, exclusive, auto_delete, arguments
        )

        yield from queue.declare(timeout)
        return queue

    @BaseChannel._ensure_channel_is_open
    @asyncio.coroutine
    def close(self) -> None:
        self.__channel.close()

        if not self._closing.done():
            self._closing.set_result(self)

        self.__channel = None

    @BaseChannel._ensure_channel_is_open
    def set_qos(self, prefetch_count: int = 0, prefetch_size: int = 0, all_channels=False, timeout: int = None):
        f = self._create_future(timeout=timeout)

        self.__channel.basic_qos(
            f.set_result,
            prefetch_count=prefetch_count,
            prefetch_size=prefetch_size,
            all_channels=all_channels
        )

        return f

    @BaseChannel._ensure_channel_is_open
    def queue_delete(self, queue_name: str, timeout: int = None,
                     if_unused: bool = False, if_empty: bool = False, nowait: bool = False):

        f = self._create_future(timeout=timeout)

        self.__channel.queue_delete(
            callback=f.set_result,
            queue=queue_name,
            if_unused=if_unused,
            if_empty=if_empty,
            nowait=nowait
        )

        return f

    @BaseChannel._ensure_channel_is_open
    def exchange_delete(self, exchange_name: str, timeout: int = None, if_unused=False, nowait=False):

        f = self._create_future(timeout=timeout)

        self.__channel.exchange_delete(
            callback=f.set_result, exchange=exchange_name, if_unused=if_unused, nowait=nowait
        )

        return f


__all__ = ('Channel',)



import json
from enum import IntEnum
from logging import getLogger
from pika import BasicProperties
from pika.channel import Channel
from contextlib import contextmanager
from .exceptions import MessageProcessError


log = getLogger(__name__)


class DeliveryMode(IntEnum):
    NOT_PERSISTENT = 1
    PERSISTENT = 2


class Message:
    __slots__ = (
        "body", "headers", "content_type", "content_encoding",
        "delivery_mode", "priority", "correlation_id", "reply_to",
        "expiration", "message_id", "timestamp", "type", "user_id", "app_id",
        "__lock"
    )

    def __init__(self, body: bytes, *, headers: dict = None, content_type: str = None,
                 content_encoding: str = None, delivery_mode: DeliveryMode = DeliveryMode.NOT_PERSISTENT,
                 priority: int = None, correlation_id=None,
                 reply_to: str = None, expiration: int = None,
                 message_id: str = None, timestamp: int = None,
                 type: str = None, user_id: str = None,
                 app_id: str = None):

        self.__lock = False
        self.body = body if isinstance(body, bytes) else bytes(body)
        self.headers = headers
        self.content_type = content_type
        self.content_encoding = content_encoding
        self.delivery_mode = DeliveryMode(int(delivery_mode)).value
        self.priority = priority
        self.correlation_id = correlation_id
        self.reply_to = reply_to
        self.expiration = int(expiration) if expiration else None
        self.message_id = message_id
        self.timestamp = timestamp
        self.type = type
        self.user_id = user_id
        self.app_id = app_id

    def info(self):
        return {
            "body_size": len(self.body) if self.body else 0,
            "headers": self.headers,
            "content_type": self.content_type,
            "content_encoding": self.content_encoding,
            "delivery_mode": self.delivery_mode,
            "priority": self.priority,
            "correlation_id": self.correlation_id,
            "reply_to": self.reply_to,
            "expiration": self.expiration,
            "message_id": self.message_id,
            "timestamp": self.timestamp,
            "type": str(self.type),
            "user_id": self.user_id,
            "app_id": self.app_id,
        }

    @property
    def locked(self):
        return self.__lock

    @property
    def properties(self):
        return BasicProperties(
            content_type=self.content_type,
            content_encoding=self.content_encoding,
            headers=self.headers,
            delivery_mode=self.delivery_mode,
            priority=self.priority,
            correlation_id=self.correlation_id,
            reply_to=self.reply_to,
            expiration=str(self.expiration) if self.expiration else None,
            message_id=self.message_id,
            timestamp=self.timestamp,
            type=self.type,
            user_id=self.user_id,
            app_id=self.app_id
        )

    def __repr__(self):
        return "{name}:{repr}".format(
            name=self.__class__.__name__,
            repr=json.dumps(self.info(), indent=1)
        )

    def __setattr__(self, key, value):
        if not key.startswith("_") and self.locked:
            raise ValueError("Message is locked")

        return super().__setattr__(key, value)

    def __iter__(self):
        return iter(self.body)

    def lock(self):
        self.__lock = True

    def __copy__(self):
        return Message(
            body=self.body, headers={k: v for k, v in self.headers.items()},
            content_encoding=self.content_encoding, content_type=self.content_type,
            delivery_mode=self.delivery_mode, priority=self.priority,
            correlation_id=self.correlation_id, reply_to=self.reply_to,
            expiration=self.expiration, message_id=self.message_id,
            timestamp=self.timestamp, type=self.type, user_id=self.user_id,
            app_id=self.app_id
        )


class IncomingMessage(Message):
    __slots__ = (
        '_loop', '__channel', 'cluster_id', 'consumer_tag',
        'delivery_tag', 'exchange', 'routing_key',
        'redelivered', '__no_ack', '__processed'
    )

    def __init__(self, channel: Channel, envelope, properties, body, no_ack: bool = False):
        self.__channel = channel
        self.__no_ack = no_ack
        self.__processed = False

        super().__init__(
            body=body,
            content_type=properties.content_type,
            content_encoding=properties.content_encoding,
            headers=properties.headers,
            delivery_mode=properties.delivery_mode,
            priority=properties.priority,
            correlation_id=properties.correlation_id,
            reply_to=properties.reply_to,
            expiration=properties.expiration,
            message_id=properties.message_id,
            timestamp=properties.timestamp,
            type=properties.type,
            user_id=properties.user_id,
            app_id=properties.app_id,
        )

        self.cluster_id = properties.cluster_id
        self.delivery_tag = envelope.delivery_tag
        self.exchange = envelope.exchange
        self.routing_key = envelope.routing_key
        self.redelivered = envelope.redelivered

    @contextmanager
    def proccess(self, requeue=False):
        try:
            yield self
            self.ack()
        except:
            self.reject(requeue=requeue)
            raise

    def ack(self):
        if self.__no_ack:
            log.warning("Can't ack message with \"no_ack\" flag")
            return False

        if self.__processed:
            raise MessageProcessError("Message already processed")
        else:
            self.__channel.basic_ack(delivery_tag=self.delivery_tag)
            self.__processed = True
            if not self.locked:
                self.lock()

    def reject(self, requeue=False):
        if self.__no_ack:
            raise TypeError('This message has "no_ack" flag.')

        if self.__processed:
            raise MessageProcessError("Message already processed")
        else:
            self.__channel.basic_reject(delivery_tag=self.delivery_tag, requeue=requeue)
            self.__processed = True
            if not self.locked:
                self.lock()


__all__ = ('Message', 'IncomingMessage')



from .connection import Connection, connect, connect_url
from .channel import Channel
from .exchange import Exchange, ExchangeType
from .message import Message, IncomingMessage
from .queue import Queue
from .exceptions import AMQPException, MessageProcessError


__all__ = (
    'connect', 'connect_url', 'Connection',
    'Channel', 'Exchange', 'Message', 'IncomingMessage', 'Queue',
    'AMQPException', 'MessageProcessError', 'ExchangeType'
)



import asyncio
import logging
from functools import partial

from pika.adapters import base_connection


LOGGER = logging.getLogger(__name__)


class IOLoopAdapter:
    __slots__ = 'loop',

    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop

    def add_timeout(self, deadline, callback_method):
        return self.loop.call_later(deadline, callback_method)

    @staticmethod
    def remove_timeout(handle: asyncio.Handle):
        return handle.cancel()

    def add_handler(self, fd, cb, event_state):
        if event_state & base_connection.BaseConnection.READ:
            self.loop.add_reader(fd, partial(cb, fd=fd, events=base_connection.BaseConnection.READ))
        if event_state & base_connection.BaseConnection.WRITE:
            self.loop.add_writer(fd, partial(cb, fd=fd, events=base_connection.BaseConnection.WRITE))

    def remove_handler(self, fd):
        self.loop.remove_reader(fd)
        self.loop.remove_writer(fd)

    def update_handler(self, fd, event_state):
        raise NotImplementedError()

    def start(self):
        return True

    def stop(self):
        return True


class AsyncioConnection(base_connection.BaseConnection):

    def __init__(self, parameters=None, on_open_callback=None,
                 on_open_error_callback=None,
                 on_close_callback=None, loop=None):

        self.sleep_counter = 0
        self.loop = loop or asyncio.get_event_loop()
        self.ioloop = IOLoopAdapter(self.loop)

        super().__init__(parameters, on_open_callback,
                         on_open_error_callback,
                         on_close_callback, self.ioloop,
                         stop_ioloop_on_close=False)

    def _adapter_connect(self):
        error = super()._adapter_connect()

        if not error:
            self.ioloop.add_handler(
                self.socket.fileno(), self._handle_events, self.event_state
            )

        return error

    def _adapter_disconnect(self):
        if self.socket:
            self.ioloop.remove_handler(self.socket.fileno())

        super()._adapter_disconnect()

    def _handle_disconnect(self):
        try:
            super()._handle_disconnect()
            super()._handle_write()
        except Exception as e:
            self._on_disconnect(-1, e)



import asyncio
from logging import getLogger
from types import FunctionType
from pika.channel import Channel
from .exchange import Exchange
from .message import IncomingMessage
from .common import BaseChannel, FutureStore

log = getLogger(__name__)


class Queue(BaseChannel):
    __slots__ = ('name', 'durable', 'exclusive',
                 'auto_delete', 'arguments',
                 '_channel', '__closing')

    def __init__(self, loop: asyncio.AbstractEventLoop, future_store: FutureStore,
                 channel: Channel, name, durable, exclusive, auto_delete, arguments):

        super().__init__(loop, future_store)

        self._channel = channel
        self.name = name
        self.durable = durable
        self.exclusive = exclusive
        self.auto_delete = auto_delete
        self.arguments = arguments

    def __str__(self):
        return "%s" % self.name

    def __repr__(self):
        return "<Queue(%s): auto_delete=%s, durable=%s, exclusive=%s, arguments=%r>" % (
            self, self.auto_delete, self.durable,
            self.exclusive, self.arguments,
        )

    @BaseChannel._ensure_channel_is_open
    def declare(self, timeout: int = None) -> asyncio.Future:
        log.debug("Declaring queue: %r", self)

        f = self._create_future(timeout)

        self._channel.queue_declare(
            f.set_result,
            self.name, durable=self.durable,
            auto_delete=self.auto_delete, arguments=self.arguments,
            exclusive=self.exclusive
        )

        return f

    @BaseChannel._ensure_channel_is_open
    def bind(self, exchange: Exchange,
             routing_key: str, *, arguments=None, timeout: int = None) -> asyncio.Future:

        log.debug(
            "Binding queue %r: exchange=%r, routing_key=%r, arguments=%r",
            self, exchange, routing_key, arguments
        )

        f = self._create_future(timeout)

        self._channel.queue_bind(
            f.set_result,
            self.name,
            exchange.name,
            routing_key=routing_key,
            arguments=arguments
        )

        return f

    @BaseChannel._ensure_channel_is_open
    def unbind(self, exchange: Exchange, routing_key: str,
               arguments: dict = None, timeout: int = None) -> asyncio.Future:

        log.debug(
            "Unbinding queue %r: exchange=%r, routing_key=%r, arguments=%r",
            self, exchange, routing_key, arguments
        )

        f = self._create_future(timeout)

        self._channel.queue_unbind(
            f.set_result,
            self.name,
            exchange.name,
            routing_key=routing_key,
            arguments=arguments
        )

        return f

    @BaseChannel._ensure_channel_is_open
    def consume(self, callback: FunctionType,
                no_ack: bool = False, exclusive: bool = False, arguments: dict = None):

        log.debug("Start to consuming queue: %r", self)

        def consumer(channel: Channel, envelope, properties, body: bytes):
            message = IncomingMessage(
                channel=channel,
                body=body,
                envelope=envelope,
                properties=properties,
                no_ack=no_ack
            )

            self.loop.create_task(asyncio.coroutine(callback)(message))

        self._channel.basic_consume(
            consumer_callback=consumer,
            queue=self.name,
            no_ack=no_ack,
            exclusive=exclusive,
            arguments=arguments
        )

    @BaseChannel._ensure_channel_is_open
    @asyncio.coroutine
    def get(self, *, no_ack=False, timeout=None) -> IncomingMessage:

        log.debug("Awaiting message from queue: %r", self)

        f = self._create_future(timeout)

        self._channel.basic_get(
            lambda *a: f.set_result(a),
            self.name,
            no_ack=no_ack,
        )

        channel, envelope, props, body = yield from f

        return IncomingMessage(
            channel,
            envelope,
            props,
            body,
            no_ack=no_ack,
        )

    @BaseChannel._ensure_channel_is_open
    def purge(self, timeout=None) -> asyncio.Future:

        log.warning("Purging queue: %r", self)

        f = self._create_future(timeout)
        self._channel.queue_purge(f.set_result, self.name)
        return f

    @BaseChannel._ensure_channel_is_open
    def delete(self, *, if_unused=True, if_empty=True, timeout=None) -> asyncio.Future:
        log.warning("Deleting %r", self)

        self._futures.reject_all(RuntimeError("Queue was deleted"))

        future = self._create_future(timeout)

        self._channel.queue_delete(
            future.set_result,
            self.name,
            if_unused=if_unused,
            if_empty=if_empty
        )

        return future


__all__ = ('Queue',)



# @copyright (c) 2002-2016 Acronis International GmbH. All rights reserved.
from setuptools import setup, find_packages


setup(
    name='aio-pika',
    version='0.5.3',
    author="Dmitry Orlov <me@mosquito.su>",
    author_email="me@mosquito.su",
    license="Apache Software License",
    description="Wrapper for the PIKA for asyncio and humans.",
    long_description=open("README.rst").read(),
    platforms="all",
    classifiers=(
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Internet',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Microsoft',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
    ),
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'asyncio',
        'shortuuid',
        'pika<0.11',
        'slimurl>=0.7.6',
    ],
)
